import builtins
import unittest
import warnings

import taxonome
from taxonome.taxa import name_selector
from taxonome.taxa.base import UncertainSpeciesError
from taxonome.taxa.author import Authority
from taxonome.taxa.name_selector import TryAnotherName
import taxonome.taxa.collection

from taxonome import Name, Taxon

class AutoInput:
    """Context manager to override the input() call. Will automatically
    return the given response."""
    def __init__(self, response=''):
        self.response = response
    
    def __enter__(self):
        self.saved_input = builtins.input
        def autoinput(prompt):
            return self.response
        builtins.input = autoinput
    
    def __exit__(self, exc_type, exc_value, traceback):
        builtins.input = self.saved_input

class HyacinthBeanTest(unittest.TestCase):
    def setUp(self):
        self.bean = taxonome.Taxon("Lablab purpureus", "(L.) Sweet")
        self.bean.othernames.add(Name("Dolichos lablab","L."))
        self.bean.othernames.add(Name("Lablab niger", "Medikus"))
        self.bean.othernames.add(Name("Vigna aristata", "Piper"))
        self.ts = taxonome.TaxonSet()
        self.ts.add(self.bean)
    
    def testSynonymy(self):
        assert self.bean.hasname("Dolichos lablab")
        assert self.bean.hasname(Name("Lablab niger", "Medikus"))
    
    def testShortauth(self):
        assert self.bean.name == Name("Lablab purpureus", "Sweet.")
        assert self.bean.name == Name("Lablab purpureus", "Sweet ex. Madeup")
        
    def testTSSynonymy(self):
        assert Name("Vigna aristata","Piper.") in self.ts
        assert self.ts["Lablab niger"]
    
    def testAccnameLookup(self):
        self.assertEqual(self.ts.get_by_accepted_name(Name("Lablab purpureus",
                                "(L.) Sweet")), self.bean)

    def test_get_by_id(self):
        name = Name("Dolichos lablab","L.")
        n, an, tid = self.ts.resolve_name(name)[0]
        self.assertEqual(self.ts.get_by_id(tid), self.bean)
    
    def test_non_auth_lookup(self):
        t1 = self.ts.select(Name("Dolichos lablab", "L."))
        self.assertIs(t1, self.bean)
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "non L."))
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "sensu non L."))
        with self.assertRaises(KeyError):
            self.ts.select(Name("Dolichos lablab", "auct. non L."))
        
        # Test selecting the other way round.
        ts2 = taxonome.TaxonSet()
        ts2.add(taxonome.Taxon("Dolichos lablab", "non L."))
        with self.assertRaises(KeyError):
            ts2.select(Name("Dolichos lablab", "L."))
        
        # This should succeed.
        ts2.select(Name("Dolichos lablab", "Vahl"), strict_authority=False)

class PoaTest(unittest.TestCase):
    def setUp(self):
        self.t1 = taxonome.Taxon("Poa annua", "L.")
        self.t2 = taxonome.Taxon("Poa infirma", "H. B. & K.")
        self.t2.othernames.add(Name("Poa annua", "Cham. & Schlecht."))
        self.ts = taxonome.TaxonSet()
        self.ts.add(self.t1)
        self.ts.add(self.t2)
        self.ts._nameselector = name_selector.TerminalNameSelector(previous_choices={})
    
    def testPreferAccName(self):
        # No authority: should pick accepted name without asking for input.
        with AutoInput("N"):
            t1 = self.ts.select("Poa annua")
        self.assertEqual(t1, self.t1)
        
        # With authority: should ask the user by default.
        with AutoInput("0"), self.assertRaisesRegex(KeyError, "0"):
            self.ts.select(Name("Poa annua", "Madeup"))
        
        # With authority: When requested, prefer the accepted name:
        with AutoInput("N"):
            t1a = self.ts.select(Name("Poa annua", "Madeup"), prefer_accepted='all')
        self.assertEqual(t1a, self.t1)
        
class LiquoriceTestBase(unittest.TestCase):
    taxon_collection_class = taxonome.TaxonSet
    taxon_collection_args = ()
    
    def setUp(self):
        self.ts = self.taxon_collection_class(*self.taxon_collection_args)
        self.liquorice = Name("Glycyrrhiza glabra", "L.")
        liquorice_t = taxonome.Taxon(self.liquorice)
        liquorice_t.othernames.add(Name("Glycyrrhiza glandulifera", "Waldst. and Kit."))
        self.ts.add(liquorice_t)
        self.ch_liquorice = Name("Glycyrrhiza uralensis", "Fisch.")
        ch_liquorice_t = taxonome.Taxon(self.ch_liquorice)
        ch_liquorice_t.othernames.add(Name("Glycyrrhiza glandulifera", "Ledeb."))
        self.ts.add(ch_liquorice_t)

class LiquoriceTest(LiquoriceTestBase):
    def testHomonym(self):
        assert self.ts.select(Name("Glycyrrhiza glabra", "L.")).hasname(Name("Glycyrrhiza glandulifera", "Waldst. & Kit."))
        homonyms = self.ts.resolve_name("Glycyrrhiza glandulifera")
        self.assertEqual(len(homonyms), 2)
        assert homonyms[0][0].match_unqualified(homonyms[1][0])
        assert not homonyms[0][1].match_unqualified(homonyms[1][1])
        with AutoInput("1"):
            self.assertIn(self.ts.select("Glycyrrhiza glandulifera"), self.ts)
    
    def testFuzzyLookup(self):
        fuzzy_match = self.ts.select("Glycyrhiza glabra")
        assert self.liquorice.match_qualified(fuzzy_match.name), fuzzy_match
        with self.assertRaises(KeyError):
            self.ts.select("Glycyrrhiza echinata")
        
        fuzzy_matches = self.ts.fuzzy_resolve_name("Glycyrhiza glandulifera")
        self.assertEqual(len(fuzzy_matches), 2)
    
    def testUpgradeSubsp(self):
        n1 = Name("Glycyrrhiza glabra subsp. glabra")
        n2 = Name("Glycyrrhiza glabra var. glabra")
        n3 = Name("Glycyrrhiza glabra subsp. spammens")
        with self.assertRaises(KeyError):
            self.ts.select(n1, upgrade_subsp="none")
        match1 = self.ts.select(n1, upgrade_subsp="nominal")
        assert match1.name.match_qualified(self.liquorice), match1
        match2 = self.ts.select(n2, upgrade_subsp="nominal")
        assert match2.name.match_qualified(self.liquorice), match2
        with self.assertRaises(KeyError):
            self.ts.select(n3, upgrade_subsp="nominal")
        match3 = self.ts.select(n3, upgrade_subsp="all")
        assert match3.name.match_qualified(self.liquorice), match3
        
        n4 = Name.from_string("Glycyrrhiza glabra L. subsp. glabra")
        match4 = self.ts.select(n4, upgrade_subsp="nominal")
        assert match4.name.match_qualified(self.liquorice), match4


class RecordingNameSelector(name_selector.NameSelector):
    lastcall = None
    
    def __call__(self, name_options, name, *args, **kwargs):
        self.lastcall = name
        return super().__call__(name_options, name, *args, **kwargs)

class LiquoriceAmbigTest(LiquoriceTestBase):
    def test_ambig_tryagain(self):
        """issue #14 - if a replacement name results in an ambiguous match
        again, it will use the terminal name selector."""
        n1 = Name("Glycyrrhiza glandulifera", "Waldst. and Kit.")
        n2 = Name("Glycyrrhiza glandulifera", "Ledeb.")
        self.ts.add(taxonome.Taxon("Glycyrrhiza fabricata", "L."))
        self.ts.add(taxonome.Taxon("Glycyrrhiza fabricata", "Regel."))
        prev_choices = {repr("Glycyrrhiza glandulifera"):
            ({(n1, self.liquorice), (n2, self.ch_liquorice)}, TryAnotherName("Glycyrrhiza fabricata"))}
        ns = RecordingNameSelector(previous_choices=prev_choices)

        try:
            with warnings.catch_warnings():
                # The ambiguous name will issue a warning, which we can ignore
                warnings.simplefilter("ignore")
                self.ts.select("Glycyrrhiza glandulifera", nameselector=ns)
        except Exception:
            pass
        
        self.assertEqual(ns.lastcall, "Glycyrrhiza fabricata")

class NameParseTest(unittest.TestCase):
    def test_parse_name(self):
        n = Name.from_string("Teramnus repens (Taub.)Baker f. ssp. gracilis (Chiov.)Verdc.")
        self.assertEqual(n.plain, "Teramnus repens subsp. gracilis")
        self.assertEqual(str(n.authority), "(Chiov.) Verdc.")
        parent = n.parent
        self.assertEqual(parent.plain, "Teramnus repens")
        # f. (filius, i.e. son) is ignored for now
        self.assertEqual(str(parent.authority), "(Taub.) Baker")
        
        # N.B. var without a .
        n = Name("Vigna radiata var sublobata", "(Roxb.) Verdc.")
        self.assertEqual(n.plain, "Vigna radiata var. sublobata")
        
        # Extra spaces
        n = Name("Vigna    radiata\tvar. \t sublobata")
        self.assertEqual(n.plain, "Vigna radiata var. sublobata")
    
    def test_subspecies(self):
        n = Name("Gorilla gorilla gorilla", "Savage, 1847")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Gorilla gorilla")
        
        n = Name.from_string("Gorilla gorilla gorilla Savage, 1847")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Gorilla gorilla")
        
        n = Name("Alloteropsis semialata subsp. eckloniana", "(Nees) Gibbs-Russ.")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Alloteropsis semialata")
        
        n = Name.from_string("Alloteropsis semialata (R. Br.) Hitchc. subsp. eckloniana (Nees) Gibbs-Russ.")
        self.assertEqual(n.rank, "subspecies")
        self.assertEqual(n.parent.plain, "Alloteropsis semialata")
        self.assertEqual(str(n.parent.authority), "(R. Br.) Hitchc.")
        self.assertEqual(str(n.authority), "(Nees) Gibbs-Russ.")
        
    def test_forma(self):
        n = Name('Muhlenbergia mexicana f. ambigua')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        
        n = Name.from_string('Muhlenbergia mexicana L. f. ambigua Foo.')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        self.assertEqual(str(n.authority), "Foo.")
        self.assertEqual(str(n.parent.authority), "L.")
        
        n = Name('Muhlenbergia mexicana forma mexicana')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        
        n = Name.from_string('Muhlenbergia mexicana L. forma mexicana Bar.')
        self.assertEqual(n.rank, 'form')
        self.assertEqual(n.parent.plain, 'Muhlenbergia mexicana')
        self.assertEqual(str(n.authority), "Bar.")
        self.assertEqual(str(n.parent.authority), "L.")
        
        # With f. for filius
        n = Name.from_string('Crotalaria dolichonyx Baker f. & Martin')
        self.assertEqual(n.rank, 'species')
        self.assertEqual(n.plain, 'Crotalaria dolichonyx')
        self.assertEqual(n.authority, Authority('Baker f. & Martin'))
    
    def test_hybrid(self):
        names = [('X Achnella caduca', 'Achnella caduca', 'genus'),
                 ('x Achnella caduca', 'Achnella caduca', 'genus'),
                 ('× Achnella caduca', 'Achnella caduca', 'genus'),
                 ('×Achnella caduca', 'Achnella caduca', 'genus'),
                 ('Calammophila X baltica', 'Calammophila baltica', 'species'),
                 ('Calammophila x baltica', 'Calammophila baltica', 'species'),
                 ('Calammophila × baltica', 'Calammophila baltica', 'species'),
                 # Using the × symbol, we allow this without a space:
                 ('Lespedeza ×patentibicolor', 'Lespedeza patentibicolor', 'species'),
                ]
        
        for raw, plain, hybrid in names:
            print(raw)
            n = Name(raw)
            self.assertEqual(n.plain, plain)
            self.assertEqual(n.hybrid, hybrid)
            
            n = Name.from_string(raw + " Bloggs")
            self.assertEqual(n.plain, plain)
            self.assertEqual(str(n.authority), "Bloggs")
            self.assertEqual(n.hybrid, hybrid)
        
    def test_genus_name(self):
        n = Name("Alloteropsis", "Presl.")
        self.assertEqual(n.plain, "Alloteropsis")
        self.assertEqual(str(n.authority), "Presl.")
    
    def test_with_year(self):
        n = Name.from_string("Balaena mysticetus Linnaeus, 1758")
        self.assertEqual(n.plain, "Balaena mysticetus")
        self.assertEqual(n.authority.year, "1758")
        
        self.assertEqual(n, Name("Balaena mysticetus", "L."))
    
    def test_invalid_name(self):
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca ovina & rubra")
        
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca sp.")
        with self.assertRaises(UncertainSpeciesError):
            Name.from_string("Festuca spp.")
        
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria sp")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria sp.")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria spp")
        with self.assertRaises(UncertainSpeciesError):
            Name("Brachiaria spp.")
    
    def test_not_capitalised(self):
        n = Name.from_string("cynodon dactylon")
        self.assertEqual(n.plain, "Cynodon dactylon")
        n = Name("echinochloa haploclada", "")
        self.assertEqual(n.plain, "Echinochloa haploclada")
    
    def test_auth_noncapital(self):
        n = Name.from_string("Trifolium pauciflorum d'Urv.")
        self.assertEqual(n.plain, "Trifolium pauciflorum")
        self.assertEqual(str(n.authority), "d'Urv.")
        
        n = Name.from_string("Trifolium pauciflorum de Madeup")
        self.assertEqual(n.plain, "Trifolium pauciflorum")
        self.assertEqual(str(n.authority), "de Madeup")


class WheatTest(unittest.TestCase):
    """See bitbucket issue 18
    
    https://bitbucket.org/taxonome/taxonome/issue/18/automatic-choice-of-the-accepted-name
    """
    def setUp(self):
        self.ts = taxonome.TaxonSet()
        self.ts.add(Taxon(Name('Triticum dicoccoides', '(Körn. ex Asch. & Graebn.) Schweinf.')))
        
        tt = Taxon(Name('Triticum timopheevii', '(Zhuk.) Zhuk.'))
        tt.othernames.add(Name('Triticum dicoccoides', 'auct.'))
        self.ts.add(tt)

    def test_resolve_name(self):
        res1 = self.ts.resolve_name('Triticum dicoccoides')
        self.assertEqual(len(res1), 2)
        
        res2 = self.ts.resolve_name(Name('Triticum dicoccoides'))
        self.assertEqual(len(res2), 2)

    def test_select(self):
        td = self.ts.select(Name('Triticum dicoccoides'), prefer_accepted='noauth')
        self.assertEqual(td.name.plain, 'Triticum dicoccoides')



from taxonome.taxa import sqlitets

from .test_taxa import LiquoriceTest

class SqliteLiquoriceTest(LiquoriceTest):
    taxon_collection_class = sqlitets.SqliteTaxonDB
    taxon_collection_args = (':memory:',)



import sqlite3
import json

from taxonome.config import config
SP_FUZZY_THRESHOLD = float(config['main']['name-fuzzy-threshold'])
from taxonome.utils import qgram
from taxonome.tracker import noop_tracker

from .author import Authority
from .base import Name, Taxon
from .collection import TaxaResource

SCHEMA = """
PRAGMA foreign_keys = ON;

create table if not exists taxa (info text, distribution text,
                                 taxonid INTEGER PRIMARY KEY);

create table if not exists names (genus text, sp_epithet text, subnames text,
                                  authority text,
                                  is_accepted integer, taxonid integer,
                                  FOREIGN KEY(taxonid) REFERENCES taxa(taxonid)
                                 );

create index if not exists spnames_ix ON names (genus, sp_epithet);
create index if not exists taxnames_ix ON names (taxonid, is_accepted);

create table if not exists individuals (info text, taxonid integer,
                                        FOREIGN KEY(taxonid) REFERENCES taxa(taxonid)
                                       );

create index if not exists taxon_individuals_ix ON individuals (taxonid);
"""

def flatten_name(name):
    """Turn a name into a tuple ready to go into the database."""
    subnames = json.dumps(name.subnames)
    authority = json.dumps(name.authority.to_dict())
    return (name.g, name.sp, subnames, authority)

def rebuild_name(namerow):
    """Build a name from a database row."""
    g, sp, subnames, authority, *extra = namerow
    subnames = json.loads(subnames)
    n = Name([g, sp] +  subnames, authority=None)
    n.authority = Authority.from_dict(json.loads(authority))
    return n

def build_unqualified_name_str(namerow):
    """Build a plain string name from a database row."""
    g, sp, subnames, authority, *extra = namerow
    subnames = json.loads(subnames)
    return " ".join([g, sp] + [r+" "+sn for (r,sn) in subnames])

class SqliteTaxonDB(TaxaResource):
    def __init__(self, dbfile):
        self.dbfile = dbfile
        # check_same_thread is not very well documented, but seems to allow us
        # to use the connection across threads.
        self.conn = sqlite3.connect(dbfile, check_same_thread=False)
        self.conn.executescript(SCHEMA)
        self.conn.commit()
    
    def add(self, taxon, commit=True):
        cur = self.conn.cursor()
        cur.execute('insert into taxa values (?, ?, NULL)',
                    (json.dumps(taxon.info), json.dumps(list(taxon.distribution))))
        
        # Sqlite docs guarantee that an INTEGER PRIMARY KEY is an alias for rowid,
        # so we can do this and save a lookup:
        taxonid = cur.lastrowid
        
        # Names
        cur.execute('insert into names values (?, ?, ?, ?, ?, ?)',
                        flatten_name(taxon.name) + (1, taxonid))
        for name in taxon.othernames:
            cur.execute('insert into names values (?, ?, ?, ?, ?, ?)',
                            flatten_name(name) + (0, taxonid))
        
        # Individuals
        for individual in taxon.individuals:
            cur.execute('insert into individuals values (?, ?)',
                        (json.dumps(individual), taxonid))

        if commit:
            self.conn.commit()
    
    def build(self, taxasource):
        for t in taxasource:
            self.add(t, commit=False)
        self.conn.commit()
    
    def resolve_name(self, name):
        if isinstance(name, Name):
            check_match = name.match_qualified
        else:
            # Unqualified name
            name = Name(name, authority=None)
            check_match = name.match_unqualified

        results = []
        for row in self.conn.execute('select * from names where genus=? and sp_epithet=?',
                                        (name.g, name.sp)):
            candidate = rebuild_name(row)
            if not check_match(candidate):
                continue
            
            taxonid = row[-1]
            if row[-2]:  # is_accepted
                candidate_acc = candidate
            else:
                accrow = self.conn.execute('select * from names where taxonid=? and is_accepted=1',
                                        (row[-1],)).fetchone()
                candidate_acc = rebuild_name(accrow)
            
            results.append((candidate, candidate_acc, taxonid))
        
        return results
    
    def __contains__(self, name):
        if isinstance(name, Taxon):
            return name.name in self
        return bool(self.resolve_name(name))

    def __len__(self):
        cur = self.conn.execute('select count(*) from taxa')
        return cur.fetchone()[0]
    
    def fuzzy_resolve_name(self, name, tracker=noop_tracker):
        if isinstance(name, Name):
            check_auth = True
            rawname = name.plain
        else:
            # Unqualified name
            rawname = name
            name = Name(name, authority=None)
            check_auth = False
        
        # Fuzzy matching - check rows where either the genus or the species matches
        target_qgrams = qgram.qgrams(rawname)
        candidates = [] # (score, row)
        for row in self.conn.execute('select * from names where genus=? or sp_epithet=?',
                                        (name.g, name.sp)):
            candidate_qgrams = qgram.qgrams(build_unqualified_name_str(row))
            score = qgram.overlap(candidate_qgrams, target_qgrams)
            if score > SP_FUZZY_THRESHOLD:
                candidates.append((score, row))
        
        if not candidates:
            return []
        
        # Only check the names with the best fuzzy match
        candidates.sort(key=lambda x: x[0], reverse=True)
        bestscore = candidates[0][0]
        
        # Check authorities, find accepted names, and build results
        results = []
        for score, row in candidates:
            if score < bestscore:
                break
            candidate = rebuild_name(row)
            if check_auth and not candidate.authority==name.authority:
                continue
            
            taxonid = row[-1]
            if row[-2]:  # is_accepted
                candidate_acc = candidate
            else:
                accrow = self.conn.execute('select * from names where taxonid=? and is_accepted=1',
                                        (row[-1],)).fetchone()
                candidate_acc = rebuild_name(accrow)
            
            results.append((candidate, candidate_acc, taxonid))
        
        return results

    def __iter__(self):
        # This is not very efficient, but it's used to get a sample of rows
        # to display in the GUI. It would also be used if someone built one
        # SqliteTaxonDB from another, but that doesn't seem too important.
        cur = self.conn.execute('select taxonid from taxa')
        for tid, in cur:
            yield self.get_by_id(tid)
    
    def get_by_id(self, taxonid):
        infoj, distribj = self.conn.execute('select info, distribution from taxa where taxonid=?', (taxonid,)).fetchone()
        namerows = self.conn.execute('select * from names where taxonid=?', (taxonid,))
        othernames = set()
        for row in namerows:
            name = rebuild_name(row)
            if row[-2]: # accepted name
                accname = name
            else:
                othernames.add(name)
        
        t = Taxon(accname)
        t.othernames = othernames
        t.info = json.loads(infoj)
        t.distribution = set(json.loads(distribj))
        
        individuals = self.conn.execute('select * from individuals where taxonid=?', (taxonid,))
        t.individuals = [json.loads(r[0]) for r in individuals]
        
        return t



import unittest

from taxonome import Name
from taxonome.taxa.name_selector import NameSelector, TerminalNameSelector

from .test_matching import VerboseTracker
from .test_taxa import AutoInput

class NameSelectorTestCase(unittest.TestCase):
    def setUp(self):
        self.liquorice = Name("Glycyrrhiza glabra", "L.")
        self.Gg1 = Gg1 = Name("Glycyrrhiza glandulifera", "Waldst. and Kit.")

        self.ch_liquorice = Name("Glycyrrhiza uralensis", "Fisch.")
        self.Gg2 = Gg2 = Name("Glycyrrhiza glandulifera", "Ledeb.")
        
        self.options = [(self.Gg1, self.liquorice, None),
                        (self.Gg2, self.ch_liquorice, None)]

        self.Gg_ambig = Name("Glycyrrhiza glandulifera", "")
        self.previous_choices = {repr(self.Gg_ambig): \
                ({(Gg1, self.liquorice), (Gg2, self.ch_liquorice)},
                 (Gg1, self.liquorice) )
            }
    
    def test_one_answer(self):
        ns = NameSelector(previous_choices={})
        option = (self.Gg1, self.liquorice, None)
        self.assertEqual(ns([option], self.Gg_ambig), option)
    
    def test_get_previous_choice(self):
        ns = NameSelector(previous_choices=self.previous_choices)

        self.assertEqual(ns(self.options, self.Gg_ambig), self.options[0])
    
    def test_store_user_choice(self):
        user_choices = {}
        ns = TerminalNameSelector(previous_choices=user_choices)
        with AutoInput("1"):
            res = ns(self.options, self.Gg_ambig)
        
        self.assertEqual(res, self.options[0])
        self.assertEqual(user_choices, self.previous_choices)
        



# -*- coding: utf-8 -*-
"""This module fetches information from the World Register of Marine Species
webservice.

http://www.marinespecies.org/

Technical note: tested using the Python 3 patches for suds available at
https://bitbucket.org/bernh/suds-python-3-patches/wiki/Home
"""
from suds.client import Client

from taxonome import Name, Taxon
from taxonome.taxa.collection import TaxaResource, TaxonSet

WSDL_URL = "http://www.marinespecies.org/aphia.php?p=soap&wsdl=1"

def _make_taxon(record):
    tax = Taxon(record.valid_name, record.valid_authority)
    
    if record.status == "accepted":
        tax.lsid = record.lsid
        tax.url = str(record.url)
    else:
        syn = Name(record.scientificname, record.authority)
        tax.othernames.add(syn)

    return tax

class WoRMSTaxaResource(TaxaResource):
    def __init__(self):
        super().__init__()
        self.service = Client(WSDL_URL)
        self.cache_by_id = {}
    
    def get_by_id(self, aphia_id):
        tax = self.cache_by_id[aphia_id]

        if tax.url is None:
            # Cached record is from a synonym - get the main record
            tax = self.cache_by_id[aphia_id] = \
                _make_taxon(self.service.service.getAphiaRecordByID(aphia_id))

        for synrecord in self.service.service.getAphiaSynonymsByID(aphia_id):
            syn = Name(synrecord.scientificname, synrecord.authority)
            tax.othernames.add(syn)
        
        return tax
    
    def resolve_name(self, key):
        if isinstance(key, Name):
            return [(n, an, tid) for n, an, tid in self.resolve_name(key.plain) \
                                            if n.authority==key.authority]
        
        names = []
        for record in self.service.service.getAphiaRecords(key):
            n = Name(record.scientificname, record.authority)
            an = Name(record.valid_name, record.valid_authority)
            self.cache_by_id[record.valid_AphiaID] = _make_taxon(record)
            names.append((n, an, record.valid_AphiaID))
        return names

_instance = None
def _get_instance():
    global _instance
    if not _instance:
        _instance = WoRMSTaxaResource()
    return _instance

def select(*args, **kwargs):
    return _get_instance().select(*args, **kwargs)

def resolve_name(*args, **kwargs):
    return _get_instance().resolve_name(*args, **kwargs)



import unittest

from taxonome import Name, Taxon, TaxonSet
from taxonome.taxa import match_taxa
from taxonome.tracker import NoopTracker

class VerboseTracker(NoopTracker):
    def name_event(self, name, event):
        print("name event", name, event)
    
    def name_transform(self, name, newname, event, **kwargs):
        print("name transform", name, newname, event, kwargs)
    
    def start_taxon(self, taxon):
        print("start taxon", taxon)
    
    def reset(self):
        print("--RESET--")
    
    def unreadable_name(self, rawname, error):
        print("unreadable name", rawname, error)

class MatchingTest(unittest.TestCase):
    def test_crotalaria(self):
        target = TaxonSet()
        target.add(Taxon(Name("Crotalaria virgulata subsp. forbesii", "(Baker) Polhill")))
        target.add(Taxon(Name("Crotalaria virgulata subsp. grantiana", "(Harv.) Polhill")))
        tomatch = [Taxon(Name("Crotalaria virgulata subsp. forbesii")),
                   Taxon(Name("Crotalaria virgulata subsp. grantiana"))
                  ]
        res = match_taxa(tomatch, target, tracker=VerboseTracker())
        self.assertEqual(len(res), 2)
